# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import pytest
from openlineage.common.provider.snowflake import fix_account_name, fix_snowflake_sqlalchemy_uri


@pytest.mark.parametrize(
    "source,target",
    [
        (
            "snowflake://user:pass@xy123456.us-east-1.aws/database/schema",
            "snowflake://xy123456.us-east-1.aws/database/schema",
        ),
        (
            "snowflake://xy123456/database/schema",
            "snowflake://xy123456.us-west-1.aws/database/schema",
        ),
        (
            "snowflake://xy12345.ap-southeast-1/database/schema",
            "snowflake://xy12345.ap-southeast-1.aws/database/schema",
        ),
        (
            "snowflake://user:pass@xy12345.south-central-us.azure/database/schema",
            "snowflake://xy12345.south-central-us.azure/database/schema",
        ),
        (
            "snowflake://user:pass@xy12345.us-east4.gcp/database/schema",
            "snowflake://xy12345.us-east4.gcp/database/schema",
        ),
        (
            "snowflake://user:pass@organization-account/database/schema",
            "snowflake://organization-account/database/schema",
        ),
        (
            "snowflake://user:p[ass@organization-account/database/schema",
            "snowflake://organization-account/database/schema",
        ),
        (
            "snowflake://user:pass@organization]-account/database/schema",
            "snowflake://organization%5D-account/database/schema",
        ),
    ],
)
def test_snowflake_sqlite_account_urls(source, target):
    assert fix_snowflake_sqlalchemy_uri(source) == target


# Unit Tests using pytest.mark.parametrize
@pytest.mark.parametrize(
    "name, expected",
    [
        ("xy12345", "xy12345.us-west-1.aws"),  # No '-' or '_' in name
        ("xy12345.us-west-1.aws", "xy12345.us-west-1.aws"),  # Already complete locator
        ("xy12345.us-west-2.gcp", "xy12345.us-west-2.gcp"),  # Already complete locator for GCP
        ("xy12345aws", "xy12345aws.us-west-1.aws"),  # AWS without '-' or '_'
        ("xy12345-aws", "xy12345-aws"),  # AWS with '-'
        ("xy12345_gcp-europe-west1", "xy12345.europe-west1.gcp"),  # GCP with '_'
        ("myaccount_gcp-asia-east1", "myaccount.asia-east1.gcp"),  # GCP with region and '_'
        ("myaccount_azure-eastus", "myaccount.eastus.azure"),  # Azure with region
        ("myorganization-1234", "myorganization-1234"),  # No change needed
        ("my.organization", "my.organization.us-west-1.aws"),  # Dot in name
    ],
)
def test_fix_account_name(name, expected):
    assert fix_account_name(name) == expected
    assert (
        fix_snowflake_sqlalchemy_uri(f"snowflake://{name}/database/schema")
        == f"snowflake://{expected}/database/schema"
    )



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0
import json
import os
import subprocess
import sys
from collections import defaultdict
from functools import cached_property
from typing import Dict, Generator, List, Optional, TextIO

from openlineage.client.event_v2 import Dataset, RunEvent, RunState
from openlineage.client.facet_v2 import (
    data_quality_assertions_dataset,
    error_message_run,
    job_type_job,
    sql_job,
)
from openlineage.client.run import InputDataset
from openlineage.client.uuid import generate_new_uuid
from openlineage.common.provider.dbt.local import DbtLocalArtifactProcessor
from openlineage.common.provider.dbt.processor import (
    DbtVersionRunFacet,
    ModelNode,
    ParentRunMetadata,
    UnsupportedDbtCommand,
)
from openlineage.common.provider.dbt.utils import (
    HANDLED_COMMANDS,
    generate_run_event,
    get_dbt_command,
    get_dbt_log_path,
    get_dbt_profiles_dir,
    get_event_timestamp,
    get_job_type,
    get_node_unique_id,
    get_parent_run_metadata,
)
from openlineage.common.utils import (
    add_command_line_args,
    add_or_replace_command_line_option,
    get_from_nullable_chain,
    get_next_lines,
    has_lines,
)


class DbtStructuredLogsProcessor(DbtLocalArtifactProcessor):
    should_raise_on_unsupported_command = True

    def __init__(
        self,
        dbt_command_line: List[str],
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.dbt_command_line: List[str] = dbt_command_line
        self.profiles_dir: str = get_dbt_profiles_dir(command=self.dbt_command_line)
        self.dbt_log_file_path: str = get_dbt_log_path(command=self.dbt_command_line)

        self.node_id_to_ol_run_id: Dict[str, str] = defaultdict(lambda: str(generate_new_uuid()))

        # sql query ids are incremented sequentially per node_id
        self.node_id_to_sql_query_id: Dict[str, Dict[str, int]] = defaultdict(lambda: {"next_id": 1})
        self.node_id_to_sql_start_event: Dict[str, RunEvent] = {}

        self.node_id_to_inputs: Dict[str, List[ModelNode]] = {}
        self.node_id_to_output: Dict[str, ModelNode] = {}

        # will be populated when some dbt events are collected
        self._compiled_manifest: Dict = {}
        self._dbt_version: Optional[str] = None
        self._dbt_log_file: Optional[TextIO] = None

    @cached_property
    def dbt_command(self) -> str:
        return get_dbt_command(self.dbt_command_line)

    @property
    def dbt_version(self) -> Optional[str]:
        """
        Extracted from the first structured log MainReportVersion
        """
        return self._dbt_version

    @cached_property
    def catalog(self) -> Optional[Dict]:
        if os.path.isfile(self.catalog_path):
            return self.load_metadata(self.catalog_path, [1], self.logger)
        return None

    @cached_property
    def profile(self):
        profile_dict = self.load_yaml_with_jinja(
            os.path.expanduser(os.path.join(self.profiles_dir, "profiles.yml"))
        )[self.profile_name]
        if not self.target:
            self.target = profile_dict["target"]

        current_profile = profile_dict["outputs"][self.target]
        return current_profile

    @property
    def compiled_manifest(self) -> Dict:
        """
        Manifest is loaded and cached.
        It's loaded when the dbt structured logs are generated and processed.
        """
        if self._compiled_manifest != {}:
            return self._compiled_manifest
        elif os.path.isfile(self.manifest_path):
            self._compiled_manifest = self.load_metadata(self.manifest_path, [2, 3, 4, 5, 6, 7], self.logger)
            return self._compiled_manifest
        else:
            return {}

    def parse(self) -> Generator[RunEvent, None, None]:
        """
        This executes the dbt command and parses the structured log events emitted.
        OL events are generated and yielded when relevant.
        """
        if self.dbt_command not in HANDLED_COMMANDS:
            raise UnsupportedDbtCommand(
                f"dbt integration for structured logs doesn't recognize dbt command "
                f"{self.dbt_command_line} - operation should be one of {HANDLED_COMMANDS}"
            )

        self.extract_adapter_type(self.profile)

        self.extract_dataset_namespace(self.profile)

        for line in self._run_dbt_command():
            ol_event = self._parse_structured_log_event(line)
            if ol_event:
                yield ol_event

    def _parse_structured_log_event(self, line: str) -> Optional[RunEvent]:
        """
        dbt generates structured events https://docs.getdbt.com/reference/events-logging
        relevant events are consumed to generate OL events.
        They are usually composed of start/finish events:
        1. NodeStart/NodeFinish for node lifecycle events
        2. MainReportVersion/CommandCompleted for dbt command lifecycle events
        3. SQLQuery/SQLQueryStatus/CatchableExceptionOnRun For SQL query lifecycle events
        """
        dbt_event = None
        try:
            dbt_event = json.loads(line)
        except ValueError:
            # log that can't be consumed
            return None

        dbt_event_name = dbt_event["info"]["name"]

        if dbt_event_name == "MainReportVersion":
            self._dbt_version = dbt_event["data"]["version"][1:]
            start_event = self._parse_dbt_start_command_event(dbt_event)
            self._setup_dbt_run_metadata(start_event)
            return start_event

        elif dbt_event_name == "CommandCompleted":
            end_event = self._parse_command_completed_event(dbt_event)
            return end_event

        if get_node_unique_id(dbt_event) is None:
            return None

        ol_event = None

        if dbt_event_name == "NodeStart":
            ol_event = self._parse_node_start_event(dbt_event)
        elif dbt_event_name == "SQLQuery":
            ol_event = self._parse_sql_query_event(dbt_event)
        elif dbt_event_name in ("SQLQueryStatus", "CatchableExceptionOnRun"):
            ol_event = self._parse_sql_query_status_event(dbt_event)
        elif dbt_event_name == "NodeFinished":
            ol_event = self.parse_node_finished_event(dbt_event)

        return ol_event

    def _parse_dbt_start_command_event(self, event) -> RunEvent:
        parent_run_metadata = get_parent_run_metadata()
        event_time = get_event_timestamp(event["info"]["ts"])
        run_facets = self.dbt_version_facet()
        if parent_run_metadata:
            run_facets["parent"] = parent_run_metadata

        start_event_run_id = str(generate_new_uuid())
        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }

        start_event = generate_run_event(
            event_type=RunState.START,
            event_time=event_time,
            run_id=start_event_run_id,
            job_name=self.job_name,
            job_namespace=self.job_namespace,
            job_facets=job_facets,
            run_facets=run_facets,
        )

        return start_event

    def _parse_node_start_event(self, event: Dict) -> RunEvent:
        node_unique_id = get_node_unique_id(event)
        run_id = self.node_id_to_ol_run_id[node_unique_id]
        node_start_time = get_event_timestamp(event["data"]["node_info"]["node_started_at"])

        parent_run_metadata = self.dbt_run_metadata.to_openlineage()
        run_facets = {"parent": parent_run_metadata, **self.dbt_version_facet()}

        job_name = self._get_job_name(event)
        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }

        inputs = [
            self.node_to_dataset(node=model_input, has_facets=True)
            for model_input in self._get_model_inputs(node_unique_id)
        ]
        outputs = [self.node_to_output_dataset(node=self._get_model_node(node_unique_id), has_facets=True)]

        return generate_run_event(
            event_type=RunState.START,
            event_time=node_start_time,
            run_id=run_id,
            run_facets=run_facets,
            job_name=job_name,
            job_namespace=self.job_namespace,
            job_facets=job_facets,
            inputs=inputs,
            outputs=outputs,
        )

    def parse_node_finished_event(self, event) -> RunEvent:
        resource_type = event["data"]["node_info"]["resource_type"]
        node_unique_id = get_node_unique_id(event)
        node_finished_at = get_event_timestamp(event["data"]["node_info"]["node_finished_at"])
        node_status = event["data"]["node_info"]["node_status"]
        run_id = self.node_id_to_ol_run_id[node_unique_id]

        parent_run_metadata = self.dbt_run_metadata.to_openlineage()
        run_facets = {"parent": parent_run_metadata, **self.dbt_version_facet()}

        job_name = self._get_job_name(event)

        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }

        event_type = RunState.COMPLETE

        if node_status not in ("success", "pass"):
            event_type = RunState.FAIL
            error_message = event["data"]["run_result"]["message"]
            run_facets["errorMessage"] = error_message_run.ErrorMessageRunFacet(
                message=error_message, programmingLanguage="sql"
            )

        inputs = [
            self.node_to_dataset(node=model_input, has_facets=True)
            for model_input in self._get_model_inputs(node_unique_id)
        ]
        outputs = [self.node_to_output_dataset(node=self._get_model_node(node_unique_id), has_facets=True)]
        if resource_type == "test":
            success = node_status == "pass"
            assertion = self._get_assertion(node_unique_id, success)
            assertion_facet = data_quality_assertions_dataset.DataQualityAssertionsDatasetFacet(
                assertions=[assertion]
            )
            attached_dataset = self._get_attached_dataset(node_unique_id)
            dataset_facets = attached_dataset.facets
            dataset_facets["dataQualityAssertions"] = assertion_facet
            inputs = [
                Dataset(
                    name=attached_dataset.name, namespace=attached_dataset.namespace, facets=dataset_facets
                )
            ]

        return generate_run_event(
            event_type=event_type,
            event_time=node_finished_at,
            run_id=run_id,
            run_facets=run_facets,
            job_name=job_name,
            job_namespace=self.job_namespace,
            job_facets=job_facets,
            inputs=inputs,
            outputs=outputs,
        )

    def _get_attached_dataset(self, test_node_id: str) -> InputDataset:
        """
        gets the input the data test is attached to
        """
        all_nodes = {**self.compiled_manifest["nodes"], **self.compiled_manifest["sources"]}
        test_node = all_nodes[test_node_id]
        attached_node_id = test_node["attached_node"]
        attached_model_node = self._get_model_node(attached_node_id)
        return self.node_to_dataset(node=attached_model_node, has_facets=True)

    def _get_assertion(self, node_id: str, success: bool) -> data_quality_assertions_dataset.Assertion:
        manifest_test_node = self.compiled_manifest["nodes"][node_id]
        name = manifest_test_node["test_metadata"]["name"]
        return data_quality_assertions_dataset.Assertion(
            assertion=name,
            success=success,
            column=get_from_nullable_chain(manifest_test_node["test_metadata"], ["kwargs", "column_name"]),
        )

    def _parse_sql_query_event(self, event) -> RunEvent:
        node_unique_id = get_node_unique_id(event)
        node_start_run_id = self.node_id_to_ol_run_id[node_unique_id]
        sql_ol_run_id = str(generate_new_uuid())
        sql_start_at = event["info"]["ts"]

        parent_run = ParentRunMetadata(
            run_id=node_start_run_id, job_name=self._get_job_name(event), job_namespace=self.job_namespace
        )

        run_facets = {"parent": parent_run.to_openlineage(), **self.dbt_version_facet()}

        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            ),
            "sql": sql_job.SQLJobFacet(query=event["data"]["sql"]),
        }

        sql_event = generate_run_event(
            event_type=RunState.START,
            event_time=sql_start_at,
            run_id=sql_ol_run_id,
            run_facets=run_facets,
            job_name=self._get_sql_job_name(event),
            job_namespace=self.job_namespace,
            job_facets=job_facets,
        )

        self.node_id_to_sql_start_event[node_unique_id] = sql_event

        return sql_event

    def _parse_sql_query_status_event(self, event):
        """
        If the sql query is successful a SQLQueryStatus is generated by dbt.
        In case of failure a CatchableExceptionOnRun is generated instead
        """
        node_unique_id = get_node_unique_id(event)
        sql_ol_run_event = self.node_id_to_sql_start_event[node_unique_id]

        dbt_node_type = event["info"]["name"]
        run_state = RunState.OTHER
        event_time = get_event_timestamp(event["info"]["ts"])
        run_facets = sql_ol_run_event.run.facets

        if dbt_node_type == "CatchableExceptionOnRun":
            run_state = RunState.FAIL
            error_message = event["data"]["exc"]
            stacktrace = event["data"]["exc_info"]
            run_facets["errorMessage"] = error_message_run.ErrorMessageRunFacet(
                message=error_message, programmingLanguage="sql", stackTrace=stacktrace
            )
        elif dbt_node_type == "SQLQueryStatus":
            run_state = RunState.COMPLETE

        return generate_run_event(
            event_type=run_state,
            event_time=event_time,
            run_id=sql_ol_run_event.run.runId,
            run_facets=run_facets,
            job_name=sql_ol_run_event.job.name,
            job_namespace=sql_ol_run_event.job.namespace,
            job_facets=sql_ol_run_event.job.facets,
        )

    def _parse_command_completed_event(self, event) -> RunEvent:
        success = event["data"]["success"]
        event_time = get_event_timestamp(event["data"]["completed_at"])
        run_facets = self.dbt_version_facet()
        if success:
            run_state = RunState.COMPLETE
        else:
            run_state = RunState.FAIL
            error_message = event["info"]["msg"]
            run_facets["errorMessage"] = error_message_run.ErrorMessageRunFacet(
                message=error_message, programmingLanguage="sql"
            )

        parent_run_metadata = get_parent_run_metadata()

        if parent_run_metadata:
            run_facets["parent"] = parent_run_metadata

        job_facets = {
            "jobType": job_type_job.JobTypeJobFacet(
                jobType=get_job_type(event),
                integration="DBT",
                processingType="BATCH",
                producer=self.producer,
            )
        }

        return generate_run_event(
            event_type=run_state,
            event_time=event_time,
            run_id=self.dbt_run_metadata.run_id,
            job_name=self.dbt_run_metadata.job_name,
            job_namespace=self.dbt_run_metadata.job_namespace,
            job_facets=job_facets,
            run_facets=run_facets,
        )

    def _setup_dbt_run_metadata(self, start_event):
        """
        The dbt command defines the parent run metadata of
        all subsequent OL events (NodeStart, NodeFinish ...)
        """
        self.dbt_run_metadata = ParentRunMetadata(
            run_id=start_event.run.runId,
            job_name=start_event.job.name,
            job_namespace=start_event.job.namespace,
        )

    def dbt_version_facet(self) -> DbtVersionRunFacet:
        return {"dbt_version": DbtVersionRunFacet(version=self.dbt_version)}

    def _get_sql_query_id(self, timestamp: str, node_id: str) -> int:
        """
        Not all adapters have the sql id defined in their dbt event.
        A node is executed by a single thread which means that sql queries of a single node are executed
        sequentially and their status is also reported sequentially.
        This function gives us a surrogate query id. It's auto-incremented.
        It gives the same id for the couple node_id, timestamp.
        """
        ids = self.node_id_to_sql_query_id[node_id]
        if timestamp not in ids:
            next_id = ids["next_id"]
            ids[timestamp] = next_id
            ids["next_id"] += 1

        return ids[timestamp]

    def _get_sql_job_name(self, event) -> str:
        """
        The name of the sql job is as follows
        {node_job_name}.sql.{incremental_id}
        """
        timestamp_str = event["info"]["ts"]
        node_unique_id = get_node_unique_id(event)
        query_id = self._get_sql_query_id(timestamp_str, node_unique_id)
        job_name = f"{node_unique_id}.sql.{query_id}"

        return job_name

    def _get_job_name(self, event) -> str:
        """
        The job name of models, snapshots, tests and others.
        """
        return get_node_unique_id(event)

    def _run_dbt_command(self) -> Generator[str, None, None]:
        """
        This executes the dbt command.
        Logs from the dbt log file are consumed in json format.
        """
        dbt_command_line = add_command_line_args(
            self.dbt_command_line,
            arg_names=["--log-format-file", "--log-level-file"],
            arg_values=["json", "debug"],
        )
        dbt_command_line = add_or_replace_command_line_option(
            dbt_command_line, option="--write-json", replace_option="--no-write-json"
        )

        process = subprocess.Popen(dbt_command_line, stdout=sys.stdout, stderr=sys.stderr, text=True)

        try:
            while process.poll() is None:
                self._open_dbt_log_file()
                if self._dbt_log_file is None:
                    continue  # wait for the log file to be created

                if has_lines(self._dbt_log_file) > 0:
                    # Load the manifest as soon as it exists
                    self.compiled_manifest

                yield from get_next_lines(self._dbt_log_file)

            if self._dbt_log_file is not None:
                yield from get_next_lines(self._dbt_log_file)

        except Exception as e:
            process.kill()
            raise e
        finally:
            if self._dbt_log_file is not None:
                self._dbt_log_file.close()

    def _open_dbt_log_file(self):
        """
        If the log file already exists when the dbt command is executed logs are appended.
        This reads all the lines on first (and only) open to get rid of those previous lines.
        """
        if os.path.exists(self.dbt_log_file_path) and self._dbt_log_file is None:
            self._dbt_log_file = open(self.dbt_log_file_path)
            while self._dbt_log_file.readlines():
                pass

    def _get_model_node(self, node_id) -> ModelNode:
        """
        Builds a ModelNode of a given node_id
        """
        all_nodes = {**self.compiled_manifest["nodes"], **self.compiled_manifest["sources"]}
        manifest_node = all_nodes[node_id]
        catalog_node = get_from_nullable_chain(self.catalog, ["nodes", node_id])
        return ModelNode(metadata_node=manifest_node, catalog_node=catalog_node)

    def _get_model_inputs(self, node_id) -> List[ModelNode]:
        """
        Builds the model's upstream inputs
        """
        if node_id in self.node_id_to_inputs:
            return self.node_id_to_inputs[node_id]

        input_node_ids = self.compiled_manifest["parent_map"][node_id]
        inputs = [self._get_model_node(input_node_id) for input_node_id in input_node_ids]
        self.node_id_to_inputs[node_id] = inputs

        return inputs

    def _get_model_output(self, node_id) -> ModelNode:
        if node_id in self.node_id_to_output:
            return self.node_id_to_output[node_id]
        model_node = self._get_model_node(node_id)
        self.node_id_to_output[node_id] = model_node

        return model_node

    def get_dbt_metadata(
        self,
    ):
        """
        Replaced by properties
        """
        pass



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

import os
from datetime import datetime
from typing import Dict, List, Optional

from openlineage.client.event_v2 import InputDataset, Job, OutputDataset, Run, RunEvent, RunState
from openlineage.common.provider.dbt.processor import ParentRunMetadata
from openlineage.common.utils import get_from_nullable_chain, parse_single_arg

__version__ = "1.27.0"
PRODUCER = f"https://github.com/OpenLineage/OpenLineage/tree/{__version__}/integration/dbt"

# for which command structured logs consumption is implemented
HANDLED_COMMANDS = ["run", "seed", "snapshot", "test", "build"]
CONSUME_STRUCTURED_LOGS_COMMAND_OPTION = "--consume-structured-logs"


def get_event_timestamp(timestamp: str):
    """
    dbt events have a discrepancy in their timestamp formats
    This converts a given timestamp string to %Y-%m-%dT%H:%M:%S.%fZ
    It returns the given timestamp if it couldn't do the conversion
    """
    output_format = "%Y-%m-%dT%H:%M:%S.%fZ"
    input_formats = ["%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%S.%f"]
    for input_format in input_formats:
        try:
            iso_timestamp = datetime.strptime(timestamp, input_format).strftime(output_format)
            return iso_timestamp
        except ValueError:
            pass  # ignore and pass to the other format

    return timestamp


def get_dbt_command(dbt_command_line: List[str]) -> Optional[str]:
    dbt_command_line_tokens = set(dbt_command_line)
    for command in HANDLED_COMMANDS:
        if command in dbt_command_line_tokens:
            return command
    return None


def generate_run_event(
    event_type: RunState,
    event_time: str,
    run_id: str,
    job_name: str,
    job_namespace: str,
    inputs: Optional[List[InputDataset]] = None,
    outputs: Optional[List[OutputDataset]] = None,
    job_facets: Optional[Dict] = None,
    run_facets: Optional[Dict] = None,
) -> RunEvent:
    inputs = inputs or []
    outputs = outputs or []
    job_facets = job_facets or {}
    run_facets = run_facets or {}
    return RunEvent(
        eventType=event_type,
        eventTime=event_time,
        run=Run(runId=run_id, facets=run_facets),
        job=Job(
            namespace=job_namespace,
            name=job_name,
            facets=job_facets,
        ),
        inputs=inputs,
        outputs=outputs,
        producer=PRODUCER,
    )


def get_dbt_profiles_dir(command: List[str]) -> str:
    """
    Based on https://docs.getdbt.com/docs/core/connect-data-platform/connection-profiles#advanced-customizing-a-profile-directory
    Gets the profiles directory
    """
    from_command = parse_single_arg(command, ["--profiles-dir"])
    from_env_var = os.getenv("DBT_PROFILES_DIR")
    default_dir = "~/.dbt/"
    return from_command or from_env_var or default_dir


def get_dbt_log_path(command: List[str]) -> str:
    """
    Based on this https://docs.getdbt.com/reference/global-configs/logs
    Gets the path of the dbt log file
    """
    project_dir = parse_single_arg(command, ["--project-dir"], default="./")
    default_log_path = os.path.join(project_dir, "logs")
    from_command = parse_single_arg(command, ["--log-path"], default=None)
    from_env_var = os.getenv("DBT_LOG_PATH")
    logs_dir = from_command or from_env_var or default_log_path
    return os.path.expanduser(os.path.join(logs_dir, "dbt.log"))


def get_parent_run_metadata():
    """
    The parent job that started the dbt command. Usually the scheduler (Airflow, ...etc)
    """
    parent_id = os.getenv("OPENLINEAGE_PARENT_ID")
    parent_run_metadata = None
    if parent_id:
        parent_namespace, parent_job_name, parent_run_id = parent_id.split("/")
        parent_run_metadata = ParentRunMetadata(
            run_id=parent_run_id, job_name=parent_job_name, job_namespace=parent_namespace
        )
    return parent_run_metadata


def get_node_unique_id(event):
    return get_from_nullable_chain(event, ["data", "node_info", "unique_id"])


def get_job_type(event) -> Optional[str]:
    """
    Gets the Run Event's job type
    """
    node_unique_id = get_node_unique_id(event)
    node_type = event["info"]["name"]
    if node_type == "SQLQuery":
        return "SQL"
    elif node_type in ("MainReportVersion", "CommandCompleted"):
        return "JOB"
    elif node_unique_id.startswith("model."):
        return "MODEL"
    elif node_unique_id.startswith("snapshot."):
        return "SNAPSHOT"
    elif node_unique_id.startswith("seed."):
        return "SEED"
    elif node_unique_id.startswith("test."):
        return "TEST"

    return None



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

GITHUB_LOCATION = (
    "https://github.com/OpenLineage/OpenLineage/tree/main/integration/common/openlineage/schema/"
)



# Copyright 2018-2025 contributors to the OpenLineage project
# SPDX-License-Identifier: Apache-2.0

from typing import ClassVar, Dict, List, Optional
from urllib.parse import urlparse

from openlineage.client.event_v2 import Dataset as OpenLineageDataset
from openlineage.client.event_v2 import InputDataset, OutputDataset
from openlineage.client.facet_v2 import (
    BaseFacet,
    DatasetFacet,
    datasource_dataset,
    documentation_dataset,
    schema_dataset,
    symlinks_dataset,
)
from openlineage.client.utils import RedactMixin
from openlineage.common.models import DbColumn, DbTableSchema


class Source(RedactMixin):
    _skip_redact: ClassVar[List[str]] = ["scheme", "name"]

    def __init__(
        self,
        scheme: Optional[str] = None,
        authority: Optional[str] = None,
        connection_url: Optional[str] = None,
        name: Optional[str] = None,
    ):
        self.scheme = scheme
        self.authority = authority
        self._name = name
        self.connection_url = connection_url

        if (scheme or authority) and name:
            raise RuntimeError("scheme + authority and namespace are exclusive options")

    def __eq__(self, other):
        return self.name == other.name and self.connection_url == other.connection_url

    def __repr__(self):
        authority = "//" + self.authority if self.authority else ""
        return f"Source({self.scheme!r}:{authority} - {self.connection_url!r})"

    @property
    def name(self) -> str:
        if self._name:
            return self._name
        if self.authority:
            return f"{self.scheme}://{self.authority}"
        return f"{self.scheme}"


class Field(RedactMixin):
    _skip_redact: ClassVar[List[str]] = ["name", "type", "tags"]

    def __init__(
        self,
        name: str,
        type: str,
        tags: Optional[List[str]] = None,
        description: Optional[str] = None,
    ):
        self.name = name
        self.type = type
        self.tags = tags
        self.description = description

        if self.tags is None:
            self.tags = []

    @staticmethod
    def from_column(column: DbColumn):
        return Field(name=column.name, type=column.type, description=column.description)

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.type == other.type
            and self.tags == other.tags
            and self.description == other.description
        )

    def __repr__(self):
        return f"Field({self.name!r},{self.type!r}, \
                       {self.tags!r},{self.description!r})"


class Dataset(RedactMixin):
    _skip_redact: ClassVar[List[str]] = ["name"]

    def __init__(
        self,
        source: Source,
        name: str,
        fields: Optional[List[Field]] = None,
        description: Optional[str] = None,
        custom_facets: Optional[Dict[str, BaseFacet]] = None,
        input_facets: Optional[Dict[str, BaseFacet]] = None,
        output_facets: Optional[Dict[str, BaseFacet]] = None,
    ):
        if fields is None:
            fields = []
        if custom_facets is None:
            custom_facets = {}
        if input_facets is None:
            input_facets = {}
        if output_facets is None:
            output_facets = {}
        self.source = source
        self.name = name
        self.fields = fields
        self.description = description
        self.custom_facets = custom_facets
        self.input_facets = input_facets
        self.output_facets = output_facets

    @staticmethod
    def from_table(
        source: Source,
        table_name: str,
        schema_name: Optional[str] = None,
        database_name: Optional[str] = None,
    ):
        return Dataset(
            name=Dataset._to_name(
                schema_name=schema_name,
                table_name=table_name,
                database_name=database_name,
            ),
            source=source,
        )

    @staticmethod
    def from_table_schema(
        source: Source,
        table_schema: DbTableSchema,
        database_name: Optional[str] = None,
        data_location: Optional[str] = None,
    ):
        parsed_path = urlparse(data_location)
        custom_facets = (
            {
                "symlinks": symlinks_dataset.SymlinksDatasetFacet(
                    identifiers=[
                        symlinks_dataset.Identifier(
                            namespace=f"{parsed_path.scheme}://{parsed_path.netloc}",  # type: ignore
                            name=str(parsed_path.path),
                            type="TABLE",
                        )
                    ]
                )
            }
            if data_location is not None
            else None
        )

        return Dataset(
            name=Dataset._to_name(
                schema_name=table_schema.schema_name,
                table_name=table_schema.table_name.name,
                database_name=database_name,
            ),
            source=source,
            fields=[
                # We want to maintain column order using ordinal position.
                Field.from_column(column)
                for column in sorted(table_schema.columns, key=lambda x: x.ordinal_position)
            ],
            custom_facets=custom_facets,  # type: ignore
        )

    @staticmethod
    def _to_name(
        table_name: str,
        schema_name: Optional[str] = None,
        database_name: Optional[str] = None,
    ):
        # Prefix the table name with database and schema name using
        # the format: {database_name}.{table_schema}.{table_name}.
        name = [table_name]
        if schema_name is not None:
            name = [schema_name] + name
        if database_name is not None:
            name = [database_name] + name

        return ".".join(name)

    def __eq__(self, other):
        return (
            self.source == other.source
            and self.name == other.name
            and self.fields == other.fields
            and self.description == other.description
        )

    def __repr__(self):
        return f"Dataset({self.source!r},{self.name!r}, \
                         {self.fields!r},{self.description!r})"

    def to_openlineage_dataset(self) -> OpenLineageDataset:
        facets: Dict[str, DatasetFacet] = {
            "dataSource": datasource_dataset.DatasourceDatasetFacet(
                name=self.source.name,
                uri=self.source.connection_url or "",
            )
        }
        if self.description:
            facets["documentation"] = documentation_dataset.DocumentationDatasetFacet(
                description=self.description
            )

        if self.fields is not None and len(self.fields):
            facets["schema"] = schema_dataset.SchemaDatasetFacet(
                fields=[
                    schema_dataset.SchemaDatasetFacetFields(field.name, field.type, field.description)
                    for field in self.fields
                ]
            )

        if self.custom_facets:
            facets.update(self.custom_facets)  # type: ignore

        if len(self.input_facets):
            return InputDataset(
                namespace=self.source.name,
                name=self.name,
                facets=facets,
                inputFacets=self.input_facets,  # type: ignore
            )
        if len(self.output_facets):
            return OutputDataset(
                namespace=self.source.name,
                name=self.name,
                facets=facets,
                outputFacets=self.output_facets,  # type: ignore
            )

        return OpenLineageDataset(namespace=self.source.name, name=self.name, facets=facets)
